from typing import Union

import pandas as pd
import numpy as np
from numpy.typing import ArrayLike
from icecream import ic


def get_fourier_coef_real(x: ArrayLike, k: int = 1, angle: float = 0.0):
    cos = np.cos(
        np.linspace(angle, k * 2 * np.pi + angle - 2 * np.pi / len(x), k * len(x))
    )
    return np.mean(cos * x)


def get_fourier_coef_imag(x: ArrayLike, k: int = 1, angle: float = 0.0):
    sin = np.sin(
        np.linspace(angle, k * 2 * np.pi + angle - 2 * np.pi / len(x), k * len(x))
    )
    return np.mean(sin * x)



import pandas as pd
import numpy as np
from typing import Union


def polyfit(ts: Union[pd.DataFrame, pd.Series], degree: int, t_offset=None) -> np.array:
    """Polynomial fit using numpy's 'polyfit'.

    Args:
        ts (Union[pd.DataFrame, pd.Series]): Values for polynomial fit (x=ts.index, y=ts.to_numpy())
        degree (int): Polynomial
        t_offset (bool, optional): time offset (x=ts.index+t_offset)

    Returns:
        np.array: Polynomial coeficients

    It was tried to implement this function using the more recent polynomial approximation as recommended by numpy as shown below. However, there were problems when 'ts' is a DataFrame.

    if not full:
      return np.polynomial.polynomial.Polynomial.fit(ts.index, ts.to_numpy(), degree).convert().coef
    else:
      return np.polynomial.polynomial.Polynomial.fit(ts.index, ts.to_numpy(), degree)
    """
    if not t_offset:
        return np.polyfit(ts.index, ts.to_numpy(), degree)
    else:
        return np.polyfit(ts.index + t_offset, ts.to_numpy(), degree)



from typing import Union

import pandas as pd
import numpy as np
from numpy.typing import ArrayLike
from scipy.ndimage import convolve1d
from scipy.optimize import minimize_scalar
from icecream import ic

from trimes.fourier import get_fourier_coef_real


def extend(x: ArrayLike, n: int, mode: str) -> Union[pd.DataFrame, pd.Series]:
    if mode == "wrap":
        if x.ndim > 1:
            return np.concatenate((x[:n, :], x), axis=0)
        else:
            return np.concatenate((x[:n], x), axis=0)


def get_angle(x: ArrayLike, **kwargs) -> float:
    f = lambda y, args: -get_fourier_coef_real(args, angle=y)
    opt_res = minimize_scalar(f, args=(x), bounds=(-np.pi, np.pi), **kwargs)
    return opt_res.x



from collections.abc import Callable

import numpy as np
from numpy.typing import ArrayLike
import pandas as pd
from icecream import ic

from trimes.base import resample


def get_metric_time_series(
    ts: pd.Series | pd.DataFrame,
    reference: pd.Series,
    metric: Callable,
    sample_time: float | ArrayLike | None = None,
    resample_ts: bool = False,
) -> float | ArrayLike:
    """Get metric (e.g. root mean squared error) of a time series.

    Serves also as an interface to the metrics of the scikit-learn package (see sklearn.metrics module).

    Args:
        ts (pd.Series | pd.DataFrame): time series.

        reference (pd.Series): reference to compare to.

        metric (Callable): A callable as defined in sklearn.metrics, see for example sklearn.metrics.root_mean_squared_error (must have same arguments).

        sample_time (float | ArrayLike | None, optional): Sample time of ts. Defaults to None (will be calculated from 'ts.index').

        resample_ts (bool, optional): If True, 'ts' wil be resampled according to 'reference'. Defaults to False (sampling time of 'ts' and 'reference' must be equal).

    Returns:
        float | ArrayLike: results for metric (for each column if 'ts' is a DataFrame)
    """
    if resample_ts:
        ts = resample(ts, reference.index.values)
    if sample_time is None:
        sample_weight = np.diff(reference.index.values)
    elif np.isscalar(sample_time):
        sample_weight = np.broadcast_to(sample_time, len(reference) - 1)
    else:
        sample_weight = sample_time
    return get_metric_np(
        ts.to_numpy()[:-1],
        reference.to_numpy()[:-1],
        metric,
        sample_weight=sample_weight,
    )


def get_metric_np(
    ts: ArrayLike,
    reference: ArrayLike,
    metric: Callable,
    sample_weight: ArrayLike | None = None,
    **kwargs,
) -> float | ArrayLike:
    """Get metric from time series data in numpy format.

    Serves also as an interface to the metrics of the scikit-learn package (see sklearn.metrics module).

    Args:
        ts (ArrayLike): time series

        reference (ArrayLike): reference to compare to.

        metric (Callable): A callable as defined in sklearn.metrics, see for example sklearn.metrics.root_mean_squared_error (must have same arguments).

        sample_weight (ArrayLike | None, optional): Sample weights (sampling time for time series). Defaults to None.

    Returns:
        float | ArrayLike: metric result
    """
    if ts.ndim > 1:
        multioutput = "raw_values"
        reference = np.broadcast_to(reference, (ts.ndim, len(sample_weight))).T
    else:
        multioutput = "uniform_average"
    return metric(
        reference, ts, sample_weight=sample_weight, multioutput=multioutput, **kwargs
    )


def integral_abs_error(
    y_true: ArrayLike,
    y_pred: ArrayLike,
    sample_weight: ArrayLike | None = None,
    multioutput="uniform_average",
) -> float | ArrayLike:
    """Integral absolute error (area).

    Metric similar to metrics defined in the scikit-learn package (see for example sklearn.metrics.root_mean_squared_error).

    Args:
        y_true (ArrayLike): true values

        y_pred (ArrayLike): predicted values

        sample_weight (_type_, optional): Sample weights. Defaults to None.

        multioutput (str, optional): See scikit-learn package. Defaults to "uniform_average".

    Returns:
        float | ArrayLike: result
    """
    weighted_error = _weighted_error(y_true, y_pred, sample_weight)
    integrated_error = np.trapezoid(np.abs(weighted_error), axis=0)
    return _handle_sklearn_multioutput(multioutput, integrated_error)


def integral_squared_error(
    y_true, y_pred, sample_weight=None, multioutput="uniform_average"
) -> float | ArrayLike:
    """Integral squared error.

    Metric similar to metrics defined in the scikit-learn package (see for example sklearn.metrics.root_mean_squared_error).

    Args:
        y_true (ArrayLike): true values

        y_pred (ArrayLike): predicted values

        sample_weight (_type_, optional): Sample weights. Defaults to None.

        multioutput (str, optional): See scikit-learn package. Defaults to "uniform_average".

    Returns:
        float | ArrayLike: result
    """
    weighted_error = _weighted_error(y_true, y_pred, sample_weight)
    integrated_error = np.trapezoid(np.square(weighted_error), axis=0)
    return _handle_sklearn_multioutput(multioutput, integrated_error)


def _weighted_error(y_true, y_pred, sample_weight):
    return (y_pred - y_true) * sample_weight[:, np.newaxis]


def _handle_sklearn_multioutput(
    multioutput: str | None | ArrayLike, metric_result: ArrayLike
):
    """See scikit-learn package and its metrics module (e.g. sklearn.metrics.root_mean_squared_error)."""
    if isinstance(multioutput, str):
        if multioutput == "raw_values":
            return metric_result
        elif multioutput == "uniform_average":
            # pass None as weights to np.average: uniform mean
            multioutput = None
    return np.average(metric_result, weights=multioutput)


def get_cosine_similarity(a: np.array, b: np.array) -> float:
    """Get cosine similarity of vectors a and b.

    Cosine similarity is a measure of similarity that does not depend on the magnitudes of the vectors, but only on their angle.
    Vectors a and b must be of same length (for example results of two variables of a time domain simulation).

    Args:
        a (np.array): vector
        b (np.array): vector

    Returns:
        float: cosine similarity
    """
    return np.dot(a, b) / (np.linalg.norm(a) * np.linalg.norm(b))


def get_cosine_similarity_matrix(a: np.array) -> np.array:
    """Get cosine similarity matrix of columns in a.

    Returns a symmetric matrix with cosine similarity values between the columns in a (e.g. indices [x,y] are the similarity between columns x and y).

    Args:
        a (np.array): matrix to calculate similarity of columns (e.g. each column is a result of a variable in time domain)

    Returns:
        np.array: similarity matrix

    TODO: use scikitlearn instead
    """
    cosine_similarity = np.zeros((a.shape[1], a.shape[1]), dtype=float)
    for col_num_1 in range(a.shape[1]):
        for col_num_2 in range(0, col_num_1):
            cosine_similarity[col_num_1, col_num_2] = get_cosine_similarity(
                a[:, col_num_1], a[:, col_num_2]
            )
            cosine_similarity[col_num_2, col_num_1] = cosine_similarity[
                col_num_1, col_num_2
            ]
    np.fill_diagonal(cosine_similarity, 1)
    return cosine_similarity


def get_cosine_similarity_matrix_as_df(a: pd.DataFrame) -> pd.DataFrame:
    """Get cosine similarity of columns in a pandas DataFrame.

    See also get_cosine_similarity_matrix. This method adds columns (headers) and indices to the similarity matrix using the columns of a.

    Args:
        a (pd.DataFrame): DataFrame with matrix to calculate similarity of columns (e.g. each column is a result of a variable in time domain)

    Returns:
        pd.DataFrame: DataFrame with similarity matrix and columns/indices from a
    """
    cosine_similarity = get_cosine_similarity_matrix(a.to_numpy())
    return pd.DataFrame(cosine_similarity, columns=a.columns, index=a.columns)



from typing import Union
from collections.abc import Iterable

import pandas as pd
import numpy as np
from numpy.typing import ArrayLike
from icecream import ic


def get_sample(
    ts: Union[pd.DataFrame, pd.Series], time: Union[np.float64, np.array]
) -> Union[pd.DataFrame, pd.Series]:
    """Get first sample after 'time'.

    Args:
        ts (Union[pd.DataFrame, pd.Series]): time series
        time (np.float64 | np.array): Point(s) in time

    Returns:
        Union[pd.DataFrame, pd.Series]: Sample(s)
    """
    return ts.iloc[np.searchsorted(ts.index.values, time)]  # uses binary search


def get_sample_shifted(
    ts: Union[pd.DataFrame, pd.Series], time: np.float64 | list[np.float64], shift: int
) -> Union[pd.DataFrame, pd.Series]:
    """Get sample after 'time' but shifted by 'shift' samples.

    Example: If shift = -1, the first sample after 'time' is searched and then the sample before that is resturned (i.e. the first sample before 'time').

    Args:
        ts (Union[pd.DataFrame, pd.Series]): time series
        time (np.float64 | list[np.float64]): point(s) in time
        shift (int): Samples are shifted by 'shift'

    Returns:
        Union[pd.DataFrame, pd.Series]: Sample(s)
    """
    return ts.iloc[np.searchsorted(ts.index.values, time) + shift]


def get_samples_around(
    ts: Union[pd.DataFrame, pd.Series],
    time: np.float64,
    num_samples_before: int = -1,
    num_samples_after: int = 0,
) -> Union[pd.DataFrame, pd.Series]:
    """Get samples around a point(s) in time (between 'num_samples_before' and 'num_samples_after', these values are relative to the first sample after 'time'). By default, the samples before and after 'time' are returned.

    Args:
        ts (Union[pd.DataFrame, pd.Series]): time series
        time (np.float64): point(s) in time
        num_samples_before (int, optional): shift before
        num_samples_after (int, optional): shift after

    Returns:
        _type_: Sample values
    """
    index = np.searchsorted(ts.index.values, time)
    return ts.iloc[index + num_samples_before : index + num_samples_after]


def get_sample_const(df: pd.DataFrame, time: list[np.float64]):
    return df.iloc[
        np.rint(
            (time - df.index.values[0]) / (df.index.values[1] - df.index.values[0])
        ).astype(int)
    ]


def interp_df(df: pd.DataFrame, time: np.array) -> pd.DataFrame:
    """Interpolate (linear) DataFrame at 'time'.

    Args:
        df (pd.DataFrame): time series
        time (np.array): point(s) in time

    Returns:
        pd.DataFrame: interpolated value(s)
    """
    return pd.DataFrame(
        interp_np_matrix(time, df.index.to_numpy(), df.to_numpy()),
        columns=df.columns,
        index=time,
    )


def interp_np_matrix(new_x: np.array, old_x: np.array, y: np.array) -> np.matrix:
    """Interpolate (linear) numpy matrix at points 'new_x'.

    Args:
      new_x (np.array): new x values
      old_x (np.array): old x values
      y (np.array): old values

    Returns:
        np.matrix: Interpolated values
    """
    interpolated = np.empty((len(new_x), y.shape[1]), dtype=np.float64)

    for y_col in range(y.shape[1]):
        interpolated[:, y_col] = np.core.multiarray.interp(new_x, old_x, y[:, y_col])
    return interpolated


def interp_series(ts: pd.Series, time: np.array) -> pd.Series:
    """Interpolate (linear) time series

    Args:
        ts (pd.Series): time series
        time (_type_): point(s) in time

    Returns:
        pd.Series: Interpolated values
    """
    return pd.Series(
        np.core.multiarray.interp(time, ts.index.values, ts.values), index=time
    )


def get_between(
    ts: Union[pd.DataFrame, pd.Series], tstart: float, tend: float
) -> Union[pd.DataFrame, pd.Series]:
    """Get values between 'tstart' and 'tend'. Returns x for 'tstart' <= x < 'tend'.

    Args:
        ts (Union[pd.DataFrame, pd.Series]): time series
        tstart (int): start time
        tend (int): end time

    Returns:
        Union[pd.DataFrame, pd.Series]: Values in range
    """
    indices = np.searchsorted(ts.index.to_numpy(), [tstart, tend])
    return ts.iloc[indices[0] : indices[1]]


def get_index(ts: Union[pd.DataFrame, pd.Series], time: np.array) -> int | ArrayLike:
    """Get first index after 'time'.

    Args:
        ts (Union[pd.DataFrame, pd.Series]): time series
        time (np.array): Point(s) in time.

    Returns:
        Union[pd.DataFrame, pd.Series]: index
    """
    return np.searchsorted(ts.index.values, time)


def resample(
    ts: Union[pd.DataFrame, pd.Series], new_time: Union[np.array, Iterable, float, int]
) -> Union[pd.DataFrame, pd.Series]:
    """Resample time series at 'new_time'.

    Args:
        ts (Union[pd.DataFrame, pd.Series]): time series
        x (Union[np.array, Iterable, float, int]): New sample time.

    Returns:
        Union[pd.DataFrame, pd.Series]: Resampled values
    """
    if not isinstance(new_time, np.ndarray):
        new_time = np.arange(ts.index.values[0], ts.index.values[-1], new_time)
    if isinstance(ts, pd.DataFrame):
        return interp_df(ts, new_time)
    else:
        return interp_series(ts, new_time)


def get_delta(ts: Union[pd.DataFrame, pd.Series], tstart: np.float64, tend: np.float64):
    """Get difference between values at 'tend' and 'tstart'

    Args:
        ts (Union[pd.DataFrame, pd.Series]): time series
        tstart (np.float64): start time
        tend (np.float64): end time

    Returns:
        _type_: Difference
    """
    return get_sample(ts, tend) - get_sample(ts, tstart)


def get_delta_shift(
    ts: Union[pd.DataFrame, pd.Series],
    time: np.float64,
    num_samples_before: int,
    num_samples_after: int,
) -> Union[pd.DataFrame, pd.Series]:
    """Get difference between samples around 'time'. The samples are shifted by 'num_samples_before'/'num_samples_after' indeces (e.g. '0' would be the first sample after 'time').

    Args:
        ts (Union[pd.DataFrame, pd.Series]): time series
        time (np.float64): point in time
        num_samples_before (int): sample shift before (usually negative)
        num_samples_after (int): sample shift after (usually positive)

    Returns:
        Union[pd.DataFrame, pd.Series]: Difference
    """

    index = np.searchsorted(ts.index.values, time)
    return ts.iloc[index + num_samples_after] - ts.iloc[index + num_samples_before]


def get_delta_around_event(
    ts: Union[pd.DataFrame, pd.Series],
    evemt_time: np.float64,
    num_samples_before: int = -2,
    num_samples_after: int = 1,
):
    """Returns the difference between time series values around an event. Uses 'get_delta_shifted' with default samples. The main reason to have this  function in addition to 'get_delta_shifted' is its descriptive name when used in the context of events in time domain simulations.
    The default values (-2 and 1) are chosen because numerical solvers sometimes behave unexpectedly around events. Hence, not the direct samples before and after the event are used, but a distance of one step is maintained.
    """
    return get_delta_shift(ts, evemt_time, num_samples_before, num_samples_after)


def get_delta_interp_df(
    df: pd.DataFrame, tstart: np.float64, tend: np.float64
) -> pd.DataFrame:
    """Get difference between interpolated (linear) values at 'tend' and 'tstart'

    Args:
        df (pd.DataFrame): time series
        tstart (np.float64): start time
        tend (np.float64): end time

    Returns:
        pd.DataFrame: Difference
    """
    values = interp_df(df, (tstart, tend))
    return values.iloc[1, :] - values.iloc[0, :]


def get_delta_interp_series(s: pd.Series, tstart: int, tend: int):
    """Get difference between interpolated (linear) values at 'tend' and 'tstart'

    Args:
        df (pd.Series): time series
        tstart (np.float64): start time
        tend (np.float64): end time

    Returns:
        pd.DataFrame: Difference
    """
    values = interp_series(s, (tstart, tend))
    return values.iloc[1] - values.iloc[0]


def get_duration(ts: pd.DataFrame | pd.Series):
    return ts.index.values[-1] - ts.index.values[0]


def create_pandas_series_or_frame_with_same_columns_and_index(
    data: np.array, original: Union[pd.DataFrame, pd.Series]
) -> Union[pd.DataFrame, pd.Series]:
    if isinstance(original, pd.DataFrame):
        return pd.DataFrame(data, columns=original.columns, index=original.index)
    else:
        return pd.Series(data, name=original.name, index=original.index)


def superpose_series(series: list[pd.Series]):
    return pd.concat(series, axis=1).sum(axis=1)


def to_numpy_array(*args):
    return [np.array(x) for x in args]



import numpy as np
from numpy.typing import ArrayLike


def get_power_power_symmetrical_components(
    u: ArrayLike, i: ArrayLike, factor: float = 3
):
    return get_active_power_symmetrical_components(
        u, i, factor=factor
    ) + 1j * get_reactive_power_symmetrical_components(u, i, factor=factor)


def get_active_power_symmetrical_components(
    u: ArrayLike, i: ArrayLike, factor: float = 3
):
    return factor * (np.dot(np.real(u), np.real(i)) + np.dot(np.imag(u), np.imag(i)))


def get_reactive_power_symmetrical_components(
    u: ArrayLike, i: ArrayLike, factor: float = 3
):
    return factor * (-np.dot(np.real(u), np.imag(i)) + np.dot(np.imag(u), np.real(i)))



from collections.abc import Callable
import operator
from functools import partial

import pandas as pd
import numpy as np
from numpy.typing import ArrayLike
from icecream import ic
import sklearn.metrics


from trimes.base import resample, get_between, get_duration
from trimes.metrics import get_metric_time_series, integral_abs_error


def subtract(
    ts: pd.Series,
    reference: pd.Series | pd.DataFrame,
    extend: bool = False,
    resample_ts: bool = False,
) -> pd.Series | pd.DataFrame:
    """Difference between 'ts' and 'reference' (subtract 'reference' from 'ts').

    Args:
        ts (pd.Series | pd.DataFrame): time series

        reference (pd.Series): reference time series

        extend (bool, optional): extend the result beyond the boundaries of 'ts' if the boundaries of 'reference' are beyond 'ts' (which is the default behavior of the 'sub' method of pandas but might not be desired). Defaults to False (i.e. the boundaries of 'reference' - and therefore also of the returned time series - are limited to the boundaries of 'ts').

        resample_ts (bool): If True, 'ts' is resampled with the index of 'reference'. If False, the sampling time of 'ts' and 'reference' must be equal. Defaults to False.

    Returns:
        pd.Series | pd.DataFrame: difference
    """
    if not extend:
        reference = _avoid_extension(ts, reference)
    if resample_ts:
        ts = resample(ts, reference.index.values)
    return ts.sub(reference, axis=0)


def add(
    ts: pd.Series,
    reference: pd.Series | pd.DataFrame,
    extend: bool = False,
    resample_ts: bool = False,
) -> pd.Series | pd.DataFrame:
    if not extend:
        reference = _avoid_extension(ts, reference)
    if resample_ts:
        ts = resample(ts, reference.index.values)
    return ts.add(reference, axis=0)


def _avoid_extension(ts, reference):
    t_start = ts.index.values[0]
    t_end = ts.index.values[-1]
    cutting_ts_is_required = False
    if reference.index.values[0] < t_start or reference.index.values[-1] > t_end:
        cutting_ts_is_required = True
    if cutting_ts_is_required:
        reference = get_between(reference, t_start, t_end)
    return reference


def apply_operator_series(
    ts: pd.Series, reference: pd.Series, operator: Callable, resample_ts: bool = False
) -> np.array:
    """Apply operator (e.g. gretaer than/less than) to series.

    Args:
        ts (pd.Series): time series

        reference (pd.Series): reference to compare to

        operator (Callable): operator (e.g. from built-in operator module)

        resample_ts (bool, optional): resample_ts (bool, optional): If True, 'ts' wil be resampled according to 'reference'. Defaults to False (sampling time of 'ts' and 'reference' must be equal).

    Returns:
        np.array: boolean
    """
    if resample_ts:
        ts = resample(ts, reference.index.values)
    return operator(ts.to_numpy(), reference.to_numpy())


def apply_operator_df(
    ts: pd.DataFrame,
    reference: pd.Series,
    operator: Callable,
    resample_ts: bool = False,
) -> np.array:
    """Apply operator (e.g. gretaer than/less than) to DataFrame. See 'apply_operator_series'."""
    if resample_ts:
        ts = resample(ts, reference.index.values)
    reference_np = reference.to_numpy()
    ts_np = ts.to_numpy()
    return np.transpose([operator(col, reference_np) for col in ts_np.T])


def greater_than_series(
    ts: pd.Series, reference: pd.Series, resample_ts: bool = False
) -> np.array:
    return apply_operator_series(ts, reference, operator.gt, resample_ts)


def less_than_series(
    ts: pd.Series, reference: pd.Series, resample_ts: bool = False
) -> np.array:
    return apply_operator_series(ts, reference, operator.lt, resample_ts)


def outside_envelope(
    ts: pd.Series, envelope: pd.DataFrame, resample_ts: bool = False
) -> np.array:
    """Get samples that are outside envelope.

    Args:
        ts (pd.Series): time series

        envelope (pd.DataFrame): first column is upper boundary, second column is lower boundary.

        resample_ts (bool, optional): If True, 'ts' wil be resampled according to 'reference'. Defaults to False (sampling time of 'ts' and 'reference' must be equal).

    Returns:
        np.array: boolean array
    """
    if resample_ts:
        ts = resample(ts, envelope.index.values)
    greater = apply_operator_series(ts, envelope.iloc[:, 0], operator.gt)
    smaller = apply_operator_series(ts, envelope.iloc[:, 1], operator.lt)
    return np.logical_or(greater, smaller)


def get_time_bool(
    b: ArrayLike,
    sample_time: float | ArrayLike,
) -> float:
    """Get time duration where 'b' is True.

    Args:
        b (ArrayLike): boolean
        sample_time (ArrayLike): sampling time

    Returns:
        float: time duration
    """
    if np.isscalar(sample_time):
        return b.sum() * sample_time
    else:
        return np.sum(np.diff(sample_time)[b[:-1]])


def comparison_series(
    ts: pd.Series,
    reference: pd.Series,
    operator_or_bool: Callable | ArrayLike | None = None,
    metric: Callable = integral_abs_error,
    sample_time: float | ArrayLike | None = None,
    resample_ts: bool = False,
) -> np.float64:
    """Compare time series to a reference (and calculate error metric).

    Args:
        ts (pd.Series | pd.DataFrame): time series.

        reference (pd.Series): reference to compare to.

        operator_or_bool (Callable, optional): Operator (applied to 'ts' and 'reference') or boolean array to select time spans.

        metric (Callable, optional): A callable as defined in sklearn.metrics, see for example sklearn.metrics.root_mean_squared_error (must have same arguments).

        sample_time (float | ArrayLike | None, optional): Sample time of ts. Defaults to None (will be calculated from 'ts.index').

        resample_ts (bool, optional): If True, 'ts' wil be resampled according to 'reference'. Defaults to False (sampling time of 'ts' and 'reference' must be equal).

    Returns:
        np.float64: results of metric
    """
    if resample_ts:
        ts = resample(ts, reference.index.values)
    if operator_or_bool is not None:
        if isinstance(operator_or_bool, Callable):
            mask = apply_operator_series(
                ts, reference, operator_or_bool, resample_ts=False
            )
        else:
            mask = operator_or_bool
        if np.any(mask):
            ts_diff = reference.copy()
            ts_diff[mask] = ts[mask]
            return get_metric_time_series(
                ts_diff, reference, metric, sample_time=sample_time
            )
        else:
            return 0.0
    else:
        return get_metric_time_series(ts, reference, metric, sample_time=sample_time)


def comparison_df(
    ts: pd.DataFrame,
    reference: pd.Series,
    operator_or_bool: Callable = None,
    metric: Callable = integral_abs_error,
    resample_ts: bool = False,
    sample_time: float | ArrayLike | None = None,
) -> ArrayLike:
    """Compare time series to a reference (and calculate error metric). See 'comparison_series'."""
    if resample_ts:
        ts = resample(ts, reference.index.values)
    if operator_or_bool is not None:
        if isinstance(operator_or_bool, Callable):
            masks = apply_operator_df(
                ts, reference, operator_or_bool, resample_ts=False
            )
        else:
            masks = operator_or_bool
        errors = np.zeros(masks.shape[1])
        for col, mask in enumerate(masks.T):
            if np.any(mask):
                ts_diff = reference.copy()
                ts_diff[mask] = ts.iloc[:, col][mask]
                errors[col] = get_metric_time_series(
                    ts_diff, reference, metric, sample_time=sample_time
                )
        return errors
    else:
        return get_metric_time_series(ts, reference, metric, sample_time=sample_time)


def envelope_comparison_series(
    ts: pd.Series,
    envelope: pd.DataFrame,
    metric: Callable = integral_abs_error,
    resample_ts: bool = False,
) -> float:
    """'comparison_series' for a reference that is an envelope.

    Args:
        ts (pd.Series): time series

        envelope (pd.DataFrame): first column is upper boundary, second column is lower boundary.

        metric (Callable, optional): see 'comparison_series'. Defaults to integral_abs_error.

        resample_ts (bool, optional): If True, 'ts' wil be resampled according to 'reference'. Defaults to False (sampling time of 'ts' and 'reference' must be equal).

    Returns:
        float: metric result
    """
    error_upper = comparison_series(
        ts, envelope.iloc[:, 0], operator.gt, metric=metric, resample_ts=resample_ts
    )
    error_lower = comparison_series(
        ts, envelope.iloc[:, 1], operator.lt, metric=metric, resample_ts=resample_ts
    )
    return error_upper + error_lower


def envelope_comparison_df(
    ts: pd.DataFrame,
    envelope: pd.DataFrame,
    metric: Callable = integral_abs_error,
    resample_ts: bool = False,
) -> ArrayLike:
    """See envelope_comparison_series (only difference: 'ts' is a DataFrame)."""
    error_upper = comparison_df(
        ts, envelope.iloc[:, 0], operator.gt, metric=metric, resample_ts=resample_ts
    )
    error_lower = comparison_df(
        ts, envelope.iloc[:, 1], operator.lt, metric=metric, resample_ts=resample_ts
    )
    return error_upper + error_lower



from collections.abc import Callable

import numpy as np
import pandas as pd
from icecream import ic


def transform_time_series(ts: pd.DataFrame, transform: Callable) -> pd.DataFrame:
    """Apply transformation to time series.

    Args:
        ts (pd.DataFrame): original
        transform (Callable): transformation

    Returns:
        pd.DataFrame: Transformed time series
    """
    ts_transformed = np.reshape(
        np.concatenate(ts.apply(transform, axis=1).to_numpy()), ts.shape
    )
    return pd.DataFrame(ts_transformed, columns=ts.columns, index=ts.index)


def abc_2_dq0(abc: np.array, phi: float) -> np.array:
    """Park transform (3-phase, amplitude invariant).

    According to mathworks.com/help/sps/ref/parktransform

    Args:
        abc (np.array): Three phase signal in natural frame
        phi (float): angle between a and d axis

    Returns:
        np.array: Signals in rotating frame
    """
    abc_2_dq0_matrix: np.array = (
        2
        / 3
        * np.array(
            [
                [np.cos(phi), np.cos(phi - 2 / 3 * np.pi), np.cos(phi + 2 / 3 * np.pi)],
                [
                    -np.sin(phi),
                    -np.sin(phi - 2 / 3 * np.pi),
                    -np.sin(phi + 2 / 3 * np.pi),
                ],
                [1 / 2, 1 / 2, 1 / 2],
            ]
        )
    )
    return np.dot(abc_2_dq0_matrix, abc)


def dq0_2_abc(dq0: np.array, phi: float) -> np.array:
    """Inverse park transform (3-phase, amplitude invariant).

    According to mathworks.com/help/sps/ref/inverseparktransform

    Args:
        dq0 (np.array): Signals in rotating frame
        phi (float): angle between a and d axis

    Returns:
        np.array: Signals in natural frame
    """
    dq0_2_abc_matrix = np.array(
        [
            [np.cos(phi), -np.sin(phi), 1],
            [
                np.cos(phi - 2 / 3 * np.pi),
                -np.sin(phi - 2 / 3 * np.pi),
                1,
            ],
            [
                np.cos(phi + 2 / 3 * np.pi),
                -np.sin(phi + 2 / 3 * np.pi),
                1,
            ],
        ]
    )
    return np.dot(dq0_2_abc_matrix, dq0)


def frobenius_matrix():
    a = np.exp(2 / 3 * np.pi * 1j)
    return 1 / 3 * np.array([[1, a, a * a], [1, a * a, a], [1, 1, 1]])


def abc_2_symmetrical_components(abc: np.array):
    return np.dot(frobenius_matrix(), abc)


def symmetrical_components_2_abc(sym_cmp: np.array):
    a = -0.5 + np.sqrt(3) / 2 * 1j
    frobenius_matrix = 1 / 3 * np.array([[1, 1, 1], [1, a * a, a], [1, a, a * a]])
    return np.dot(frobenius_matrix, sym_cmp)



from typing import Union

import pandas as pd
import numpy as np
from scipy.ndimage import convolve1d
from scipy.optimize import minimize_scalar
from icecream import ic

from trimes.base import create_pandas_series_or_frame_with_same_columns_and_index
from trimes.signal_processing import extend
from trimes.filter import pt1


def rms_rolling_2(
    ts: Union[pd.DataFrame, pd.Series],
    samples_per_window: int,
    mode: str = "wrap",
) -> Union[pd.DataFrame, pd.Series]:
    """

    From https://stackoverflow.com/questions/8245687/numpy-root-mean-squared-rms-smoothing-of-a-signal

    Args:
        ts (Union[pd.DataFrame, pd.Series]): _description_
        samples_per_window (int): _description_

    Returns:
        Union[pd.DataFrame, pd.Series]: _description_
    """
    ts_squared = np.power(ts.to_numpy(), 2)
    window_averaging_weights = np.ones(samples_per_window) / float(samples_per_window)
    rms = np.sqrt(convolve1d(ts_squared, window_averaging_weights, mode=mode, axis=0))
    return create_pandas_series_or_frame_with_same_columns_and_index(rms, ts)


def rms_rolling(
    ts: Union[pd.DataFrame, pd.Series],
    samples_per_window: int,
    normalize_magnitude: bool = False,
) -> Union[pd.DataFrame, pd.Series]:

    normalization_factor = np.sqrt(2) if normalize_magnitude else 1
    ts_squared = np.cumsum(np.power(ts.to_numpy(), 2), axis=0)
    func_rms = (
        lambda x: np.sqrt(
            (x[samples_per_window:] - x[:-samples_per_window]) / samples_per_window
        )
        * normalization_factor
    )
    rms = np.apply_along_axis(func_rms, 0, ts_squared)
    rms = extend(rms, samples_per_window, "wrap")
    return create_pandas_series_or_frame_with_same_columns_and_index(rms, ts)


def rms_min_max(
    ts: Union[pd.DataFrame, pd.Series],
    filter_time_constant: float,
) -> Union[pd.DataFrame, pd.Series]:
    if isinstance(ts, pd.Series):
        ts = ts.to_frame()
    rms = np.zeros(ts.shape[0], dtype=float)

    factor = np.pi / (3 * np.sqrt(3))

    first_row = ts.to_numpy()[0, :]
    rms[0] = factor * (np.max(first_row) - np.min(first_row))

    for n, row_values in enumerate(ts.to_numpy()[1:, :], 1):
        val = factor * (np.max(row_values) - np.min(row_values))
        rms[n] = rms[n - 1] + (val - rms[n - 1]) / filter_time_constant * (
            ts.index.values[n] - ts.index.values[n - 1]
        )
    return pd.DataFrame(rms, index=ts.index)



import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.axes._axes import Axes
from cycler import cycler
from icecream import ic

from trimes.base import resample


def plot_2y(
    ts_axis1: pd.DataFrame | pd.Series,
    ts_axis2: pd.DataFrame | pd.Series,
    kwargs_ax1: dict = {},
    kwargs_ax2: dict = {},
    colors: list | None = None,
) -> tuple[Axes, Axes]:
    """Plot with two y-axes (i.e. a secondary y-axis on the right).

    Args:
        ts_axis1 (pd.DataFrame | pd.Series): time series for left axis
        ts_axis2 (pd.DataFrame | pd.Series): time series for right axis
        kwargs_ax1 (dict, optional): kwargs for left axis. Defaults to {}.
        kwargs_ax2 (dict, optional): kwargs for right axis. Defaults to {}.
        colors (list | None, optional): colors used. Defaults to None (default colormap is used).

    Returns:
        tuple[Axes, Axes]: left and right axes
    """
    ts_axis1.plot(**kwargs_ax1)
    ax1 = plt.gca()
    plt.legend(loc="upper left")

    ax2 = ax1.twinx()
    if colors is None:
        colors = get_pyplot_default_colormap_hex()
    color_index = ts_axis1.shape[1] % len(colors)
    color_cycler = cycler(color=colors[color_index:] + colors[:color_index])
    ax2.set_prop_cycle(color_cycler)
    ts_axis2.plot(ax=ax2, **kwargs_ax2)
    plt.legend(loc="upper right")
    return ax1, ax2


def fill_between(
    ts1: pd.Series,
    ts2: pd.Series,
    ax: Axes | None = None,
    resample_ts: bool = True,
    **kwargs,
) -> None:
    """Fill area between time series 'ts1' and 'ts2'.

    Args:
        ax (Axes): matpotlib axes
        ts1 (pd.Series): time series
        ts2 (pd.Series): time series
        resample_ts (bool): If True, 'ts2' is resampled according to the index of 'ts1'. Default is True.
        kwargs: keyword arguments for 'ax.fill_between'
    """
    if resample_ts:
        ts2 = resample(ts2, ts1.index.values)
    if not ax:
        ax = plt.gca()
    ax.fill_between(ts1.index.values, ts1.to_numpy(), ts2.to_numpy(), **kwargs)


def add_vertical_line_to_plot(x: float | int, **kwargs):
    plt.axvline(
        x=x,  # vertical line
        **kwargs,
    )


def add_point_to_plot(
    x: float, y: float, label: str = "", ha="center", va="top", **kwargs
):
    """Add point to plot.

    Args:
        x (float): x coordinate
        y (float): y coordinate
        label (str, optional): Defaults to "".
        ha: 'center', 'right', 'left'
        va: 'top', 'bottom', 'center', 'baseline', 'center_baseline'
    """
    ax = plt.gca()
    ax.plot(x, y, "o", **kwargs)
    if label:
        plt.gca().text(x, y, label, ha=ha, va=va)


def get_pyplot_default_colormap_hex():
    return plt.rcParams["axes.prop_cycle"].by_key()["color"]


def get_pyplot_default_color_cycler(start_index=0):
    cmap = get_pyplot_default_colormap_hex()
    return cycler(color=cmap[start_index:] + cmap[:start_index])
